/*
 *  This file is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  It is part of httppc, which is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <netinet/in.h> // ntohs(), htons()
#include <errno.h>      // errno
#include <string.h>     // strncpy(), memset()
#include <stdio.h>      // snprintf(), sscanf()
#include <stdarg.h>     // va_start(), va_end()

#include "config.h"
#include "httppc.h"
#include "internal.h"

// receive from sockfd line and write into buffer dst
static int
recv_line(int sockfd, int timeout, char *dst, int len)
{
    char ch;
    int count = 0;

    while (1)
    {
        int n;

        n = wait_canread(sockfd, timeout);

        if (n == 0) // timeout
        {
            logerr(PFX"timeout\n");
            errno = ETIMEDOUT;
            return -1;
        } else if (n < 0) // error
        {
            logerrno(PFX"error while waiting data from proxy"); 
            return -1;
        }

        n = recv(sockfd, &ch, 1, 0);

        if (n == 0)
        {
            logerr(PFX"recv() from proxy: EOF\n");
            errno = ETIMEDOUT;
            return -1;
        } else if (n < 0)
        {
            logerrno(PFX"recv() from proxy");
            return -1;
        }

        if (ch == '\012') // line ending is 13 10 (\015 \012)
            break;

        dst[count++] = ch;
        if (count > len)
        {
            logerr(PFX"received line from proxy is too long\n");
            return -1;
        }
    }
    dst[count] = '\0';
    return 0;
}

// connect to host through proxy, main function
static int 
proxy_connect(int sockfd, const char *serv_name, unsigned short serv_port, int in_family,
              const httppc_options_t *options, const char *http_head_lines)
{
    // hostinfo variables
    struct hostent hostinfo_res_buf, *hostinfo = 0;
    char hostinfo_buf[BUFF_LEN];
    int r, herr;
    // --
    char rw_buf[RWBUFF_LEN];    // send/recv buffer
    int bpos, wlen;             // position and the number of written bytes in buffer
    int codenum;                // proxy response number
    int httpvn1, httpvn2;       // minor/major numbers of HTTP version in proxy response

    log(PFX "connecting to proxy %s:%d...\n", options->http_proxy,
        options->http_proxy_port);

    // get proxy host information
#ifdef HAVE_GETHOSTBYNAME_R
    if (options->real_gethostbyname_r)
    {
        // Linux
# if defined(__GLIBC__)
        r = options->real_gethostbyname_r(options->http_proxy,
                                          &hostinfo_res_buf, hostinfo_buf,
                                          BUFF_LEN, &hostinfo, &herr);
        if (r)
        {
            logerr(PFX "unknown host %s\n", options->http_proxy);
            return -1;
        }
        //
# else
        hostinfo =
            options->real_gethostbyname_r(options->http_proxy,
                                          &hostinfo_res_buf, hostinfo_buf,
                                          BUFF_LEN, &herr);
        if (!hostinfo)
        {
            logerr(PFX "unknown host %s\n", options->http_proxy);
            return -1;
        }
# endif
    }
#endif
    // non reentrant gethostbyname()
    if (!hostinfo)
    {
        if (options->real_gethostbyname)
            hostinfo = options->real_gethostbyname(options->http_proxy);
        else
        {
            logerr(PFX "specify real_gethostbyname_r or real_gethostbyname values in options parameter\n");
            return -1;
        }
    }

    if (!hostinfo)
    {
        logerr(PFX "unknown host %s\n", options->http_proxy);
        return -1;
    }
    log_hostinfo(hostinfo);
    // fill http proxy address structure
    if (in_family == AF_INET)
    {
        struct sockaddr_in hp_addr; // http proxy address structure

        memset(&hp_addr, 0, sizeof(hp_addr));
        hp_addr.sin_family = in_family;
        hp_addr.sin_port = htons(options->http_proxy_port);
        hp_addr.sin_addr = *(struct in_addr *) hostinfo->h_addr;

        // connecting
        r = options->real_connect(sockfd, (struct sockaddr *) &hp_addr,
                                  sizeof(hp_addr));
#ifdef AF_INET6
    } else if (in_family == AF_INET6)
    {
        struct sockaddr_in6 hp_addr6; // http proxy address structure
        struct in_addr haddr = *(struct in_addr *) hostinfo->h_addr;

        memset(&hp_addr6, 0, sizeof(hp_addr6));
        hp_addr6.sin6_family = in_family;
        hp_addr6.sin6_port = htons(options->http_proxy_port);
        hp_addr6.sin6_addr.s6_addr[10] = hp_addr6.sin6_addr.s6_addr[11] =
            0xff;
        hp_addr6.sin6_addr.s6_addr[12] = haddr.s_addr;
        hp_addr6.sin6_addr.s6_addr[13] = haddr.s_addr >> 8;
        hp_addr6.sin6_addr.s6_addr[14] = haddr.s_addr >> 16;
        hp_addr6.sin6_addr.s6_addr[15] = haddr.s_addr >> 24;

        // connecting
        r = options->real_connect(sockfd, (struct sockaddr *) &hp_addr6,
                                  sizeof(hp_addr6));
#endif // AF_INET6
    } else
    {
        errno = EAFNOSUPPORT;
        r = -1;
    }

    if (r < 0 && (errno == EINPROGRESS || errno == EWOULDBLOCK))
        r = wait_canwrite(sockfd, options->timeout);

    if (r < 0)
    {
        logerrno(PFX "unable to connect to proxy");
        return -1;
    }
    log(PFX "connection to proxy established\n");

    // write header
    bpos = wlen =
        snprintf(rw_buf, RWBUFF_LEN, "CONNECT %s:%d HTTP/" HTTP_VERSION "\n",
                 serv_name, serv_port);
    if (wlen >= 0 && http_head_lines)
        bpos += (wlen =
                 snprintf(rw_buf + bpos, RWBUFF_LEN - bpos, http_head_lines));
    // write terminating newline
    if (wlen >= 0)
        wlen = snprintf(rw_buf + bpos, RWBUFF_LEN - bpos, "\n");
    // was there any overflow?
    if (wlen < 0)
    {
        logerr(PFX "buffer overflow in vproxy_connect()\n");
        return -1;
    }
    // send request to proxy
    log(PFX "<- %s", rw_buf);
    if (send(sockfd, rw_buf, strlen(rw_buf), 0) < 0)
    {
        logerrno(PFX "unable to send data to proxy");
        return -1;
    }
    // receive resulting string
    if (recv_line(sockfd, options->timeout, rw_buf, RWBUFF_LEN) < 0)
        return -1;
    // received, getting result code
    log(PFX "-> %s\n", rw_buf);
    if (!sscanf(rw_buf, "HTTP/%d.%d %d", &httpvn1, &httpvn2, &codenum))
    {
        logerr(PFX "unable to parse received line\n");
        return -1;
    }
    // eat up '\n'
    if (recv_line(sockfd, options->timeout, rw_buf, RWBUFF_LEN) < 0)
        return -1;
    // Successful 2xx
    if ((codenum / 100) != 2)
    {
        //    logerr(PFX"proxy does not support http tunnelling?\n");
        switch (codenum / 100)
        {
            case 5:            // Server Error 5xx
                errno = ECONNREFUSED;
                break;
            default:
                errno = ENETUNREACH;
                break;
        }
        return -1;
    }

    return 0;
}

// wrap function for connect to sockaddr structure
int
httppc_connect(int sockfd, const struct sockaddr *serv_addr, int addrlen,
               const httppc_options_t *options)
{
    char serv_name[BUFF_LEN];        // target address in IPv4 compatible notation
    unsigned short serv_port;        // target port

    // get target hostname and port from serv_addr
    if (serv_addr->sa_family == AF_INET)
    {
        const struct sockaddr_in *saddr = (const struct sockaddr_in *)serv_addr;
        serv_port = ntohs(saddr->sin_port);
        ipv4n_to_ipv4p(&saddr->sin_addr, serv_name, BUFF_LEN);
#ifdef AF_INET6
    } else if (serv_addr->sa_family == AF_INET6)
    {
        const struct sockaddr_in6 *saddr6 = (const struct sockaddr_in6 *)serv_addr;
        serv_port = ntohs(saddr6->sin6_port);
        ipv6n_to_ipv4p(&saddr6->sin6_addr, serv_name, BUFF_LEN);
#endif // AF_INET6
    } else
    {
        errno = EAFNOSUPPORT;
        return -1;
    }
    return httppc_connect2(sockfd, serv_name, serv_port, serv_addr->sa_family, options);
}

// wrap function for connect to name and port
int 
httppc_connect2(int sockfd, const char *serv_name, unsigned short serv_port, int in_family,
                const httppc_options_t *options)
{
    char buffer[BUFF_LEN];
    int bpos = 0, wlen = 0;

    log(PFX"target host is: %s:%d\n", serv_name, serv_port);
    // nocache
    if (options->args & httppc_nocache)
        bpos += (wlen = snprintf(buffer, BUFF_LEN, "Pragma: no-cache\nCache-Control: no-cache\n"));
    // keepalive
    if (wlen >= 0 && options->args & httppc_keepalive)
        bpos += (wlen = snprintf(buffer+bpos, BUFF_LEN-bpos, "Connection: Keep-Alive\nProxy-Connection: Keep-Alive\n"));
    // authorization
    if (wlen >= 0 && options->args & httppc_auth && options->proxy_user && options->proxy_pass)
    {
        char user_pass[BUFF_LEN];
        wlen = snprintf(user_pass, BUFF_LEN, "%s:%s", options->proxy_user, options->proxy_pass);
        if (wlen > 0)
        {
            char enc[BUFF_LEN*2];
            base64(enc, user_pass, wlen);
            bpos += (wlen = snprintf(buffer+bpos, BUFF_LEN-bpos, "Proxy-authorization: Basic %s\n", enc));
        }
    }
    // useragent
    if (wlen >= 0 && options->args & httppc_useragent && options->useragent)
        wlen = snprintf(buffer+bpos, BUFF_LEN-bpos, "User-Agent: %s\n", options->useragent);
    // was there any overflow?
    if (wlen < 0)
    {
        logerr(PFX"buffer overflow in httppc_connect2_args()\n");
        return -1;
    }
    return proxy_connect(sockfd, serv_name, serv_port, in_family, options, buffer);
}

/*int 
httppc_connect_ipv4(int sockfd, const char *serv_name, unsigned short serv_port,
                const char *http_proxy, unsigned short http_proxy_port,
                httppc_connect_t real_connect, httppc_gethostbyname_r_t real_gethostbyname_r,
                const char *http_head_lines, ...)
{
    va_list argptr;
    int r;
    va_start(argptr, http_head_lines);
    r = proxy_connect(sockfd, serv_name, serv_port, AF_INET, 
                      http_proxy, http_proxy_port,
                      real_connect, real_gethostbyname_r,
                      http_head_lines, argptr);
    va_end(argptr);
    return r;
}

int 
httppc_connect_ipv6(int sockfd, const char *serv_name, unsigned short serv_port,
                const char *http_proxy, unsigned short http_proxy_port,
                httppc_connect_t real_connect, httppc_gethostbyname_r_t real_gethostbyname_r,
                const char *http_head_lines, ...)
{
#ifdef AF_INET6
    va_list argptr;
    int r;
    va_start(argptr, http_head_lines);
    r = proxy_connect(sockfd, serv_name, serv_port, AF_INET6, 
                      http_proxy, http_proxy_port,
                      real_connect, real_gethostbyname_r,
                      http_head_lines, argptr);
    va_end(argptr);
    return r;
#else // !AF_INET6
    errno = EAFNOSUPPORT;
    return -1;
#endif // !AF_INET6
}
*/
