/*
 *  This file is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  It is part of httppc, which is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <netinet/in.h> // ntohs()
#include <arpa/inet.h>  // inet_aton()
#include <unistd.h>     // close(), dup2()
#include <fcntl.h>      // fcntl(), F_GETFL, F_SETFL, O_NONBLOCK
#include <errno.h>      // errno
#include <string.h>     // memcpy(), strlen(), memset(), strcpy()
#include <stdlib.h>     // malloc()
#include "config.h"
#include "httppc.h"
#include "internal.h"
#include "options.h"

//#define USE_THREADS 1

#ifdef USE_THREADS
# include <pthread.h>
#endif

// static data
static int initialized = 0;
static struct options cfg;
// --

static int
reopen_socket(int sockfd)
{
    int sock_stats;
    int newsock;

    // reopening socket, is there other way to flush socket state? FIXME
    newsock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (newsock < 0 || dup2(newsock, sockfd) < 0 || close(newsock) < 0)
    {
        logerrno("unable to reopen socket");
        return -1;
    }
    // socket was nonblocking, so made new socket as is
    sock_stats = fcntl(sockfd, F_GETFL, 0);
    if (sock_stats >= 0)
        fcntl(sockfd, F_SETFL, sock_stats | O_NONBLOCK);
    if (sock_stats < 0)
    {
        logerrno("unable to unblock reopened socket");
        return -1;
    }
    return 0;
}

#ifdef USE_THREADS
typedef struct thread_data_s
{
    int sockfd;
    struct sockaddr saddr;
    int addrlen;
    pthread_t tid;
    struct thread_data_s *prev, *next;
} thread_data_s;

static thread_data_s *thread_data_head = 0;
static pthread_mutex_t thread_data_mutex = PTHREAD_MUTEX_INITIALIZER;

static thread_data_s *
find_thread_data(int sockfd)
{
    thread_data_s *p = thread_data_head;
    while (p && p->sockfd != sockfd)
        p = p->next;
    return p;
}

static void *
connect_thread(thread_data_s *p)
{
    httppc_connect_args(p->sockfd, &p->saddr, p->addrlen,
                        cfg.http_proxy,cfg. http_proxy_port,
                        cfg.real_connect, cfg.real_gethostbyname_r, 
                        cfg.timeout,
                        cfg.conn_args, cfg.proxy_user, cfg.proxy_pass,
                        cfg.useragent);
    pthread_mutex_lock(&thread_data_mutex);
    if (p->prev)
        p->prev->next = p->next;
    else 
        thread_data_head = p->next;
    if (p->next)
        p->next->prev = p->prev;
    pthread_mutex_unlock(&thread_data_mutex);
    free(p);
    return 0;
}

static int                
threaded_connect(int sockfd, const struct sockaddr *saddr, int addrlen)
{
    thread_data_s *p;
    pthread_attr_t attr;

    pthread_mutex_lock(&thread_data_mutex);

    if (find_thread_data(sockfd))
    {
        pthread_mutex_unlock(&thread_data_mutex);
        errno = EALREADY;
        return -1;
    }

    p = (thread_data_s *) malloc(sizeof(thread_data_s));
    if (!p)
    {
        logerrno(PFX"malloc() out of memory");
        errno = ENOMEM;
        return -1;
    }
    p->sockfd = sockfd;
    p->saddr = *saddr;
    p->addrlen = addrlen;
    p->prev = 0;
    p->next = thread_data_head;
    thread_data_head = p;

    pthread_mutex_unlock(&thread_data_mutex);

    pthread_attr_init(&attr);
    pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
    pthread_create(&p->tid, &attr, &connect_thread, p);
    pthread_attr_destroy(&attr);

    errno = EINPROGRESS;
    return -1;
}
#endif /* USE_THREADS */

// replacement for system connect()
int
connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
    int res = -1;
//    struct sockaddr_in target_addr;     // used for redirection
    const struct sockaddr_in *saddr = (const struct sockaddr_in *)serv_addr;
    struct in_addr taddr;               // target address
    int sopt, soptlen = sizeof(sopt);

    init(&initialized, &cfg);
    
    log(PFX"-- connect(%d, ...)\n", sockfd);

    if (VERBOSE)
    {
        unsigned short port;
        char dot_name[BUFF_LEN];            // target address dot name
        if (ipn_to_ipv4p(serv_addr, dot_name, BUFF_LEN))
        {
            port = ntohs(saddr->sin_port);
            log(PFX"connecting to: %s, %s:%d\n", family_s(serv_addr->sa_family), dot_name, port);
        }
    }

    if (getsockopt(sockfd, SOL_SOCKET, SO_TYPE, &sopt, &soptlen) < 0)
    {
        logerrno(PFX"getsockopt()");
        return -1;
    }

    if (sopt != SOCK_STREAM)
    {
        log(PFX"socket is not SOCK_STREAM: %d(%s)\n", sopt, socktype_s(sopt));
        return cfg.real_connect(sockfd, serv_addr, addrlen);
    }

#ifdef AF_INET6
    if (!(saddr && (serv_addr->sa_family == AF_INET || serv_addr->sa_family == AF_INET6)))
    {
        log(PFX"family is neither AF_INET, nor AF_INET6: %s\n", family_s(serv_addr->sa_family));
        return cfg.real_connect(sockfd, serv_addr, addrlen);
    }
#else // !AF_INET6
    if (!(saddr && serv_addr->sa_family == AF_INET))
    {
        log(PFX"family is not AF_INET: %s\n", family_s(serv_addr->sa_family));
        return cfg.real_connect(sockfd, serv_addr, addrlen);
    }
#endif // !AF_INET6

    if (cfg.extern_lsz || cfg.local_lsz)
        ipn_to_ipv4n(serv_addr, &taddr);

    if (cfg.extern_lsz)
    {
        int ind = in_nets_list(taddr, cfg.extern_ne, cfg.extern_lsz);
        if (ind >= 0)
        {
            log(PFX"external network match: %s\n", cfg.extern_nets[ind]);
            res = -1;
            errno = ENETUNREACH;
        } else
        {
            //log(PFX"no matches in external network list, back to system connect()\n");
            res = cfg.real_connect(sockfd, serv_addr, addrlen);
            return res;
        }
    } else if (cfg.local_lsz)
    {
        int ind = in_nets_list(taddr, cfg.local_ne, cfg.local_lsz);
        if (ind >= 0)
        {
            log(PFX"local network match: %s\n", cfg.local_nets[ind]);
            res = cfg.real_connect(sockfd, serv_addr, addrlen);
            return res;
        } else
        {
            //log(PFX"no matches in local network list, connecting via proxy\n");
            res = -1;
            errno = ENETUNREACH;
        }
    } else
    {
        // neither xnets, nor lnets specified, try to figure ot our own
        res = cfg.real_connect(sockfd, serv_addr, addrlen);
        if (res < 0 && errno == EINPROGRESS)
        {
            res = wait_canrw(sockfd, cfg.timeout);
            // in case POLLERR, errno is set to EHOSTUNREACH
            if (res < 0 && errno == EHOSTUNREACH)
                if (reopen_socket(sockfd) < 0)
                    return -1;
        }
    }

#if 0
    {
        unsigned short port = ntohs(((struct sockaddr_in*)serv_addr)->sin_port);
        struct in_addr ldns_addr;
        if (inet_pton(AF_INET, "176.16.4.107", &ldns_addr) > 0
            && port == 53
            && ldns_addr.s_addr == ((struct sockaddr_in*)serv_addr)->sin_addr.s_addr)
        {
            struct hostent *hostinfo;
            hostinfo = gethostbyname(extern_dns);
            if (hostinfo)
            {
                log(PFX"redirect?\n");
                reopen_socket(sockfd);
                ((struct sockaddr_in*)serv_addr)->sin_addr = *(struct in_addr*)hostinfo->h_addr;
                memset(&target_addr, 0, sizeof(target_addr));
                target_addr.sin_family = AF_INET;
                target_addr.sin_port = htons(53);
                target_addr.sin_addr = *(struct in_addr*)hostinfo->h_addr;
                serv_addr = (struct sockaddr *)&target_addr;
                res = -1;
                errno = ENETUNREACH;
                do_real_call = 0;
            }
        }
    }
#endif

    if (res < 0 && (errno == ENETUNREACH 
                    || errno == EHOSTUNREACH
                    || errno == EINPROGRESS 
                    || errno == EWOULDBLOCK
                    || errno == ETIMEDOUT))
    {
        const httppc_options_t o = 
        {
            cfg.real_connect, cfg.real_gethostbyname, cfg.real_gethostbyname_r,
            cfg.http_proxy, cfg.http_proxy_port,
            cfg.proxy_user, cfg.proxy_pass,
            cfg.useragent,
            cfg.timeout,
            cfg.conn_args
        };

#ifdef USE_THREADS
        {
            int sockfd_stats;
            sockfd_stats = fcntl(sockfd, F_GETFL, 0);
            log_fd_stats(sockfd, sockfd_stats);
            if (sockfd_stats < 0)
            {
                logerrno(PFX"fcntl() error");
                return -1;
            }
            if (sockfd_stats & O_NONBLOCK)
                return threaded_connect(sockfd, serv_addr, addrlen);
        }
#endif
        res = httppc_connect(sockfd, serv_addr, addrlen, &o);
    }

    return res;
}

// Linux
#if defined(__GLIBC__)
#ifdef HAVE_GETHOSTBYNAME_R
int
gethostbyname_r(const char *name, struct hostent *result_buf,
                char *buf, size_t buflen,
                struct hostent **result, int *h_errnop)
{
    int r;
    init(&initialized, &cfg);
    log(PFX"-- gethostbyname_r(\"%s\")\n", name);

    if (cfg.conn_args & httppc_ldns)
    {
        log(PFX"trying real_gethostbyname_r()...\n");
        r = cfg.real_gethostbyname_r(name, result_buf, buf, buflen, result, h_errnop);

        if (r != 0)
        {
            log(PFX"... real_gethostbyname_r() failed\n");
            if (*h_errnop == HOST_NOT_FOUND)    // Unknown host 
            {
                *result = httppc_gethostbyname_r(name, result_buf, buf, buflen, h_errnop, cfg.extern_dns);
                r = *result ? 0 : -1;
            }
        } else
            log(PFX"... local dns lookup succeed\n");
    } else 
    {
        *result = httppc_gethostbyname_r(name, result_buf, buf, buflen, h_errnop, cfg.extern_dns);
        r = *result ? 0 : -1;
    }

    log_hostinfo(*result);
    return r;
}
#endif // HAVE_GETHOSTBYNAME_R
#ifdef HAVE_GETHOSTBYADDR_R
int 
gethostbyaddr_r(GETHOSTBYADDR_R_FV_TYPE addr, socklen_t len, int type, struct hostent *result_buf,
                char *buf, size_t buflen, struct hostent **result, int *h_errnop)
{
    int r;
    init(&initialized, &cfg);
    log(PFX"-- gethostbyaddr_r()\n");

    if (cfg.conn_args & httppc_ldns)
    {
        log(PFX"trying real_gethostbyaddr_r()...\n");
        r = cfg.real_gethostbyaddr_r(addr, len, type, result_buf, buf, buflen, result, h_errnop);

        if (r != 0)
        {
            log(PFX"... real_gethostbyaddr_r() failed\n");
            if (*h_errnop == HOST_NOT_FOUND)    // Unknown host 
            {
                *result = httppc_gethostbyaddr_r(addr, len, type, result_buf, buf, buflen, h_errnop, cfg.extern_dns);
                r = *result ? 0 : -1;
            }
        } else
            log(PFX"... local dns lookup succeed\n");
    } else 
    {
        *result = httppc_gethostbyaddr_r(addr, len, type, result_buf, buf, buflen, h_errnop, cfg.extern_dns);
        r = *result ? 0 : -1;
    }

    log_hostinfo(*result);
    return r;
}
#endif // HAVE_GETHOSTBYADDR_R
#else // Solaris
#ifdef HAVE_GETHOSTBYNAME_R
struct hostent *
gethostbyname_r(const char *name, struct hostent *result_buf, 
                char *buf, int buflen, int *h_errnop)
{
    struct hostent *result;
    init(&initialized, &cfg);
    log(PFX"-- gethostbyname_r(\"%s\")\n", name);

    if (cfg.conn_args & httppc_ldns)
    {
        log(PFX"trying real_gethostbyname_r()...\n");
        result = cfg.real_gethostbyname_r(name, result_buf, buf, buflen, h_errnop);

        if (!result)
        {
            log(PFX"... real_gethostbyname_r() failed\n");
            if (*h_errnop == HOST_NOT_FOUND)    // Unknown host 
                result = httppc_gethostbyname_r(name, result_buf, buf, buflen, h_errnop, cfg.extern_dns);
        } else
            log(PFX"... local dns lookup succeed\n");
    } else
        result = httppc_gethostbyname_r(name, result_buf, buf, buflen, h_errnop, cfg.extern_dns);

    log_hostinfo(result);
    return result;
}
#endif // HAVE_GETHOSTBYNAME_R
#ifdef HAVE_GETHOSTBYADDR_R
struct hostent *
gethostbyaddr_r(const char *addr, int len, int type, 
                struct hostent *result_buf,
                char *buf, int buflen, int *h_errnop)
{
    struct hostent *result;
    init(&initialized, &cfg);
    log(PFX "-- gethostbyaddr_r()\n");

    if (cfg.conn_args & httppc_ldns)
    {
        log(PFX"trying real_gethostbyaddr_r()...\n");
        result = cfg.real_gethostbyaddr_r(addr, len, type, result_buf, buf, buflen, h_errnop);

        if (!result)
        {
            log(PFX"... real_gethostbyaddr_r() failed\n");
            if (*h_errnop == HOST_NOT_FOUND)    // Unknown host 
                result = httppc_gethostbyaddr_r(addr, len, type, result_buf, buf, buflen, h_errnop, cfg.extern_dns);
        } else
            log(PFX"... local dns lookup succeed\n");
    } else
        result = httppc_gethostbyaddr_r(addr, len, type, result_buf, buf, buflen, h_errnop, cfg.extern_dns);

    log_hostinfo(result);
    return result;
}
#endif // HAVE_GETHOSTBYADDR_R
#endif

// replacement for system gethostbyname()
struct hostent *
gethostbyname(const char *name)
{
    struct hostent *result;
    init(&initialized, &cfg);
    log(PFX"-- gethostbyname(\"%s\")\n", name);

    if (cfg.conn_args & httppc_ldns)
    {
        log(PFX"trying real_gethostbyname()...\n");
        result = cfg.real_gethostbyname(name);

        if (!result)
        {
            log(PFX"... real_gethostbyname() failed\n");
            if (h_errno == HOST_NOT_FOUND)    // Unknown host 
                result = httppc_gethostbyname(name, cfg.extern_dns);
        } else
            log(PFX"... local dns lookup succeed\n");
    } else
        result = httppc_gethostbyname(name, cfg.extern_dns);

    log_hostinfo(result);
    return result;
}

#if defined(__GLIBC__)
struct hostent *
gethostbyaddr(GETHOSTBYADDR_FV_TYPE addr, socklen_t len, int type)
#else // !__GLIBC__
struct hostent *
gethostbyaddr(const char *addr, int len, int type)
#endif // !__GLIBC__
{
    struct hostent *result;
    init(&initialized, &cfg);
    log(PFX "-- gethostbyaddr()\n");

    if (cfg.conn_args & httppc_ldns)
    {
        log(PFX"trying real_gethostbyaddr()...\n");
        result = cfg.real_gethostbyaddr(addr, len, type);

        if (!result)
        {
            log(PFX"... real_gethostbyaddr() failed\n");
            if (h_errno == HOST_NOT_FOUND)    // Unknown host 
                result = httppc_gethostbyaddr(addr, len, type, cfg.extern_dns);
        } else
            log(PFX"... local dns lookup succeed\n");
    } else
        result = httppc_gethostbyaddr(addr, len, type, cfg.extern_dns);

    log_hostinfo(result);
    return result;
}

int 
socket(int domain, int type, int pnum)
{
    int res;

    init(&initialized, &cfg);

    res = cfg.real_socket(domain, type, pnum);

    log(PFX"-- socket(%d(%s), %d(%s), %d(%s)) = %d\n",
        domain, family_s(domain), type, socktype_s(type),
        pnum, protocol_s(pnum), res);

    return res;
}

#if 0
int 
select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout)
{
    int res;

    init(&initialized, &cfg);

    res = cfg.real_select(n, readfds, writefds, exceptfds, timeout);

    log(PFX"-- select() = %d\n", res);

    return res;
}

int 
poll(struct pollfd *ufds, unsigned int nfds, int timeout)
{
    int res;

    init(&initialized, &cfg);

    res = cfg.real_poll(ufds, nfds, timeout);

    log(PFX"-- poll() = %d\n", res);

    return res;
}
#endif

#ifdef HAVE_GETADDRINFO
int 
getaddrinfo(const char *node, const char *service,
            const struct addrinfo *hints, struct addrinfo **res)
{
    int r = EAI_NODATA;
    
    init(&initialized, &cfg);
    log(PFX"-- getaddrinfo(\"%s\", \"%s\", ...)\n", node ? node : "null",
        service ? service : "null");

    if (cfg.conn_args & httppc_ldns)
    {
        log(PFX"trying real_gethostbyaddr()...\n");
        r = cfg.real_getaddrinfo(node, service, hints, res);
        if (r)
            log(PFX"real_getaddrinfo(): %d: %s\n", r, gai_strerror(r));
    }

    if ((r == EAI_NONAME || r == EAI_NODATA) && node
        && (hints->ai_family == AF_UNSPEC 
         || hints->ai_family == AF_INET
         || hints->ai_family == 0))
        // The node or service is not known.  
        // This error is also returned if both node and service are NULL.
    {
        struct addrinfo *base, req;

        memcpy(&req, hints, sizeof(req));
        req.ai_family = AF_INET;
        r = cfg.real_getaddrinfo("127.0.0.1", service, &req, &base);

        if (!r)
        {
            struct hostent *hostinfo, result_buf;
            char buf[BUFF_LEN];
            int h_err;
            struct in_addr taddr;
            int dotted = inet_aton(node, &taddr);

            log(PFX"got addrinfo for localhost\n");
            log_addrinfo(base);

            log(PFX"requesting DNS for `%s'\n", node);

            if (dotted)
                hostinfo = httppc_gethostbyaddr_r(&taddr, sizeof(taddr), AF_INET,
                                                  &result_buf, buf, 
                                                  BUFF_LEN, &h_err, cfg.extern_dns);
            else
                hostinfo = httppc_gethostbyname_r(node,
                                                  &result_buf, buf, 
                                                  BUFF_LEN, &h_err, cfg.extern_dns);

            if (hostinfo || dotted)
            {
                struct addrinfo *cur = base;
                struct addrinfo **pai = res;
                int socklen = sizeof(struct sockaddr_in);
                int namelen = 0;
                const char *c = 0;
                struct in_addr *ptaddr = 0;

                if (hostinfo)
                {
                    log(PFX"found host:\n");
                    log_hostinfo(hostinfo);

                    if (hints->ai_flags & AI_CANONNAME)
                    {
                        c = hostinfo->h_name;
                        namelen = strlen(c) + 1;
                    }
                    ptaddr = (struct in_addr *)hostinfo->h_addr_list[0];
                } else if (dotted)
                {
                    log(PFX"can't find requested host, but address is dotted\n");

                    if (hints->ai_flags & AI_CANONNAME)
                    {
                        c = node;
                        namelen = strlen(node) + 1;
                    }
                    ptaddr = &taddr;
                }

                // the following code taken from glibc sources
                // modified to copy information for localhost and 
                // replace localhost address and canonical name with 
                // target hostname
                for (cur = base; cur; cur = cur->ai_next)
                {
                    struct sockaddr_in *sinp;

                    *pai = (struct addrinfo *)malloc(sizeof(struct addrinfo) + socklen + namelen);
                    if (!*pai)
                        return -EAI_MEMORY;

                    (*pai)->ai_flags = cur->ai_flags;
                    (*pai)->ai_family = AF_INET;
                    (*pai)->ai_socktype = cur->ai_socktype;
                    (*pai)->ai_protocol = cur->ai_protocol;
                    (*pai)->ai_addrlen = socklen;
                    (*pai)->ai_addr = (void *) (*pai) + sizeof(struct addrinfo);
#if SALEN
                    (*pai)->ai_addr->sa_len = socklen;
#endif /* SALEN */
                    (*pai)->ai_addr->sa_family = AF_INET;

                    sinp    = (struct sockaddr_in *)(*pai)->ai_addr;
                    memcpy (&sinp->sin_addr, ptaddr, sizeof (struct in_addr));
                    sinp->sin_port = ((struct sockaddr_in *)cur->ai_addr)->sin_port;
                    memset (sinp->sin_zero, 0, sizeof (sinp->sin_zero));

                    if (c)
                    {
                        (*pai)->ai_canonname = ((void *) (*pai) +
                                                sizeof (struct addrinfo) + socklen);
                        strcpy ((*pai)->ai_canonname, c);
                    }
                    else
                        (*pai)->ai_canonname = NULL;

                    (*pai)->ai_next = NULL;
                    pai = &((*pai)->ai_next);
                }
                r = 0;
            } else
            {
                log(PFX"can't find requested host\n");
                r = EAI_NONAME;
            }

            freeaddrinfo(base);
        }
    }
    if (!r)
        log_addrinfo(*res);
    return r;
}
#endif // HAVE_GETADDRINFO

