/*
 *  This file is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  It is part of httppc, which is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <errno.h>      // errno
#include <string.h>     // strncpy()
#include <stdlib.h>     // free()
#include <adns.h>

#include "config.h"
#include "httppc.h"
#include "internal.h"

static int
hostent_set(struct hostent *res, char *buf, int buflen, 
            const char *name, struct in_addr *addr)
{
    char *ptr = buf;
    // reset
    memset(res, 0, sizeof(*res));
    res->h_addrtype = AF_INET;
    res->h_length = sizeof(struct in_addr);

    // put address into buffer
    *(struct in_addr *)ptr = *addr;
    ptr += sizeof(struct in_addr);

    // put pointer to address into buffer
    *(struct in_addr **)ptr = (struct in_addr *)buf;
    res->h_addr_list = (char **)ptr;
    ptr += sizeof(struct in_addr*);

    // next pointer to address is NULL
    *(struct in_addr **)ptr = 0;
    ptr += sizeof(struct in_addr*);

    // pointer to aliases is NULL
    *(char **)ptr = 0;
    res->h_aliases = (char **)ptr;
    ptr += sizeof(char *);

    // put dot name into buffer
    res->h_name = ptr;
    strncpy(ptr, name, buflen-(ptr-buf));
    return 0;
}

static adns_answer *
request_addr(adns_state adns, const char *name)
{
    STRERROR_V;
    adns_query query;
    adns_answer *answer;
    int r;

    log(PFX"submiting query addr for %s\n", name);
    r = adns_submit(adns, name, adns_r_a, adns_qf_usevc, 0, &query);
    if (r)
    {
        logerrno(PFX"adns_submit()");
        return 0;
    }

    log(PFX"waiting for answer\n");
    r = adns_wait(adns, &query, &answer, 0);
    if (r)
    {
        logerr(PFX"adns_wait(): %s\n", STRERROR(r));
        return 0;
    }

    log(PFX"answer status: %s, type: %d, rrsz: %d\n", 
        adns_strerror(answer->status), answer->type, answer->rrsz);

    return answer;
}

static adns_answer *
request_name(adns_state adns, struct in_addr *addr)
{
    STRERROR_V;
    struct sockaddr_in saddr;
    adns_query query;
    adns_answer *answer;
    int r;

    memset(&saddr, 0, sizeof(saddr));
    saddr.sin_family = AF_INET;
    saddr.sin_addr = *addr;

    log(PFX"submiting query name for address 0x%x\n", saddr.sin_addr);
    r = adns_submit_reverse(adns, (struct sockaddr *)&saddr, adns_r_ptr_raw, adns_qf_usevc, 0, &query);
    if (r)
    {
        logerrno(PFX"adns_submit_reverse()");
        return 0;
    }

    log(PFX"waiting for answer\n");
    r = adns_wait(adns, &query, &answer, 0);
    if (r)
    {
        logerr(PFX"adns_wait(): %s\n", STRERROR(r));
        return 0;
    }

    log(PFX"answer status: %s, type: %d, rrsz: %d\n", 
        adns_strerror(answer->status), answer->type, answer->rrsz);

    return answer;
}

static char *
get_result(adns_answer *answer)
{
    STRERROR_V;
    char *res = 0;
    if (answer && answer->status == adns_s_ok)
    {
        int r = adns_rr_info(answer->type, 0, 0, 0, answer->rrs.bytes, &res);
        if (r)
        {
            logerr(PFX"adns_rr_info(): %s\n", STRERROR(r));
            return 0;
        }
    }
    return res;
}

static int
init_adns(adns_state *adns, const char *server)
{
    STRERROR_V;
    int r;
    char configtext[BUFF_LEN];
    adns_initflags flags = adns_qf_cname_loose;

    if (VERBOSE)
        flags |= adns_if_debug;

    log(PFX"initializing adns engine\n");
    if (snprintf(configtext, BUFF_LEN, "nameserver %s", server) < 0)
    {
        logerr(PFX"buffer overflow in init_adns()\n");
        return -1;
    }
    r = adns_init_strcfg(adns, flags, stderr, configtext);
    if (r)
    {
        logerr(PFX"adns_init_strcfg(): %s\n", STRERROR(r));
        return -1;
    }
    return 0;
}

// do hostname lookup via adns library (simple interface , tcp connection)
// thread safe version
struct hostent * 
httppc_gethostbyname(const char *name, const char *server)
{
	static struct hostent result_buf;
	static char buffer[BUFF_LEN];
	return httppc_gethostbyname_r(name, &result_buf, buffer, BUFF_LEN, 
								  &h_errno, server);
}

struct hostent *
httppc_gethostbyname_r(const char *name, struct hostent *result_buf, 
					   char *buf, int buflen, int *h_errnop,
					   const char *server)
{
    struct hostent *result = 0;
    adns_state adns;
    adns_answer *answer;

    *h_errnop = HOST_NOT_FOUND;
    
    if (init_adns(&adns, server) < 0)
        return 0;

    answer = request_addr(adns, name);

    while (answer && answer->status == adns_s_prohibitedcname && answer->cname)
        answer = request_addr(adns, answer->cname);

    if (answer && answer->status == adns_s_ok)
    {
        struct in_addr addr = *answer->rrs.inaddr;

        log(PFX"from adns: 0x%x\n", addr);
        hostent_set(result_buf, buf, buflen, name, &addr);

        result = result_buf;
        *h_errnop = 0;
    }

    adns_finish(adns);
    return result;
}

struct hostent *
httppc_gethostbyaddr(const void *addr, int len, int type,
                     const char *server)
{
    static struct hostent result_buf;
    static char buffer[BUFF_LEN];
    return httppc_gethostbyaddr_r(addr, len, type, &result_buf, buffer,
                                  BUFF_LEN, &h_errno, server);
}

#if 1
struct hostent *
httppc_gethostbyaddr_r(const void *addr, int len, int type,
                       struct hostent *result_buf,
                       char *buf, int buflen, int *h_errnop,
                       const char *server)
{
    struct hostent *result = 0;
    adns_state adns;
    adns_answer *answer;
    char *name;

    if (type != AF_INET)
    {
        errno = EAFNOSUPPORT;
        *h_errnop = NETDB_INTERNAL;
        return 0;
    }

    *h_errnop = HOST_NOT_FOUND;
    
    if (init_adns(&adns, server) < 0)
        return 0;

    answer = request_name(adns, (struct in_addr*)addr);
    name = get_result(answer);

    if (name)
    {
        log(PFX"from adns: %s\n", name);

        hostent_set(result_buf, buf, buflen, name, (struct in_addr *)addr);
        free(name);

        result = result_buf;
        *h_errnop = 0;
    }

    adns_finish(adns);
    return result;
}
#else
struct hostent *
httppc_gethostbyaddr_r(const void *addr, int len, int type,
                       struct hostent *result_buf,
                       char *buf, int buflen, int *h_errnop,
                       const char *server)
{
    struct hostent *result = 0;
    struct sockaddr_in saddr;
    char *ptr = buf;
    char dot_name[BUFF_LEN];

    *h_errnop = HOST_NOT_FOUND;
    
    if (!inet_ntop(type, addr, dot_name, BUFF_LEN))
    {
        logerrno(PFX"inet_ntop()");
        return 0;
    }

    // reset
    memset(result_buf, 0, sizeof(*result_buf));
    result_buf->h_addrtype = AF_INET;
    result_buf->h_length = sizeof(struct in_addr);

    // put address into buffer
    *(struct in_addr *)ptr = saddr.sin_addr;
    ptr += sizeof(struct in_addr);

    // put pointer to address into buffer
    *(struct in_addr **)ptr = (struct in_addr *)buf;
    result_buf->h_addr_list = (char **)ptr;
    ptr += sizeof(struct in_addr*);

    // next pointer to address is NULL
    *(struct in_addr **)ptr = 0;
    ptr += sizeof(struct in_addr*);

    // put name into buffer
    result_buf->h_name = ptr;
    strncpy(ptr, dot_name, buflen-(ptr-buf));

    result = result_buf;
    *h_errnop = 0;

    return result;
}
#endif
