/*
 *  This file is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  It is part of httppc, which is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <stdio.h>      // fprintf
#include <netinet/in.h> // ntohs()
#include <sys/socket.h> // struct sockaddr
#include <arpa/inet.h>  // inet_ntoa()
#include <sys/poll.h>   // poll events
#include <netdb.h>      // struct hostent
#include <stdarg.h>     // va_*
#include <errno.h>      // errno
#include <string.h>     // strerror()
#include <fcntl.h>

#include "config.h"
#include "internal.h"

int httppc__mode = MODE_NONE;

void
log(const char *format, ...)
{
    if (VERBOSE)
    {
        va_list argptr;
        va_start(argptr, format);
        vfprintf(stderr, format, argptr);
        va_end(argptr);
    }
}

void
logerr(const char *format, ...)
{
    if (!QUIET)
    {
        va_list argptr;
        va_start(argptr, format);
        vfprintf(stderr, format, argptr);
        va_end(argptr);
    }
}

void
logerrno(const char *format, ...)
{
    if (!QUIET)
    {
        STRERROR_V;
        va_list argptr;
        va_start(argptr, format);
        vfprintf(stderr, format, argptr);
        va_end(argptr);
        fprintf(stderr, ": %s\n", STRERROR(errno));
    }
}

void
log_sockaddr(const struct sockaddr *saddr)
{
    STRERROR_V;
    char nstr[BUFF_LEN];
    const char *res = 0;

    if (!(saddr && VERBOSE))
        return;
        
    log(PFX"Sockaddr:\n");
    log(PFX"\tfamily: %d(%s)\n", saddr->sa_family, family_s(saddr->sa_family));
    if (saddr->sa_family == AF_INET)
    {
        const struct sockaddr_in *p = (const struct sockaddr_in *)saddr;
        log(PFX"\tport: %d\n", ntohs(p->sin_port));
        res = inet_ntop(AF_INET, (void *)&p->sin_addr, nstr, BUFF_LEN);
#ifdef AF_INET6
    } else if (saddr->sa_family == AF_INET6)
    {
        const struct sockaddr_in6 *p = (const struct sockaddr_in6 *)saddr;
        log(PFX"\tport: %d\n", p->sin6_port);
        log(PFX"\tflowinfo: %d\n", p->sin6_flowinfo);
//        log(PFX"\tscope_id: %d\n", p->sin6_scope_id);
        res = inet_ntop(AF_INET6, (void *)&p->sin6_addr, nstr, BUFF_LEN);
#endif // AF_INET6
    }
    log(PFX"\taddress: %s\n", res ? res : STRERROR(errno));
}

#ifdef HAVE_GETADDRINFO
void
log_addrinfo(const struct addrinfo *ptr)
{
    int i = 0;
    while (ptr)
    {
        log(PFX"Record %d\n", ++i);
//        log(PFX"flags: 0x%x\n", ptr->ai_flags);
        log(PFX"\tfamily: %d(%s)\n", ptr->ai_family, family_s(ptr->ai_family));
        log(PFX"\tsocktype: %d(%s)\n", ptr->ai_socktype, socktype_s(ptr->ai_socktype));
        log(PFX"\tprotocol: %d(%s)\n", ptr->ai_protocol, protocol_s(ptr->ai_protocol));
        log(PFX"\tcanonical name: %s\n", ptr->ai_canonname ? ptr->ai_canonname : "unknown");
        log_sockaddr(ptr->ai_addr);
        ptr = ptr->ai_next;
    }
}
#endif // HAVE_GETADDRINFO

void
log_hostinfo(const struct hostent *hostinfo)
{
    int n;
    struct in_addr **addrl;

    if (!(hostinfo && VERBOSE))
        return;
        
//    log(PFX"from real gethostbyname():\n");
    log(PFX"\thost name: %s\n", hostinfo->h_name);
    log(PFX"\thost aliases:");
    n = 0;
    if (hostinfo->h_aliases)
        while (hostinfo->h_aliases[n])
            log(" %s", hostinfo->h_aliases[n++]);
    log("\n");
    log(PFX"\taddresses:");
    n = 0;
    addrl = (struct in_addr **)hostinfo->h_addr_list;
    if (addrl)
        while (addrl[n])
        {
            struct in_addr addr = *addrl[n++];
            const char *str = inet_ntoa(addr);
            log(" %s(0x%x)", str ? str : "invalid", addr);
        }
    log("\n");
}

void
log_poll_events(short events)
{
    if (!VERBOSE)
        return;

    log(PFX"poll() revents: 0x%x:", events);
    if (events & POLLIN)
        log(" POLLIN");
    if (events & POLLPRI)
        log(" POLLPRI");
    if (events & POLLOUT)
        log(" POLLOUT");
    if (events & POLLERR)
        log(" POLLERR");
    if (events & POLLHUP)
        log(" POLLHUP");
    if (events & POLLNVAL)
        log(" POLLNVAL");
    log("\n");
}

#ifdef USE_THREADS
void
log_fd_stats(int fd, int stats)
{
    if (!VERBOSE)
        return;

    log(PFX"socket %d flags: 0x%x:", fd, stats);
    if (stats & O_ACCMODE)
        log(" O_ACCMODE");
    if (stats & O_RDONLY)
        log(" O_RDONLY");
    if (stats & O_WRONLY)
        log(" O_WRONLY");
    if (stats & O_RDWR)
        log(" O_RDWR");
    if (stats & O_APPEND)
        log(" O_APPEND");
    if (stats & O_NONBLOCK)
        log(" O_NONBLOCK");
    if (stats & O_SYNC)
        log(" O_SYNC");
    log("\n");
}
#endif

const char * family_linux[] = {
    "AF_UNSPEC",                /* 0: Unspecified.  */
    "AF_LOCAL",                 /* 1: Local to host (pipes and file-domain).  */
    "AF_INET",                  /* 2: IP protocol family.  */
    "AF_AX25",                  /* 3: Amateur Radio AX.25.  */
    "AF_IPX",                   /* 4: Novell Internet Protocol.  */
    "AF_APPLETALK",             /* 5: Appletalk DDP.  */
    "AF_NETROM",                /* 6: Amateur radio NetROM.  */
    "AF_BRIDGE",                /* 7: Multiprotocol bridge.  */
    "AF_ATMPVC",                /* 8: ATM PVCs.  */
    "AF_X25",                   /* 9: Reserved for X.25 project.  */
    "AF_INET6",                 /* 10: IP version 6.  */
    "AF_ROSE",                  /* 11: Amateur Radio X.25 PLP.  */
    "AF_DECnet",                /* 12: Reserved for DECnet project.  */
    "AF_NETBEUI",               /* 13: Reserved for 802.2LLC project.  */
    "AF_SECURITY",              /* 14: Security callback pseudo AF.  */
    "AF_KEY",                   /* 15: PF_KEY key management API.  */
    "",                         /* 16: */
    "AF_PACKET",                /* 17: Packet family.  */
    "AF_ASH",                   /* 18: Ash.  */
    "AF_ECONET",                /* 19: Acorn Econet.  */
    "AF_ATMSVC",                /* 20: ATM SVCs.  */
    "",                         /* 21: */
    "AF_SNA",                   /* 22: Linux SNA Project */
    "AF_IRDA",                  /* 23: IRDA sockets.  */
    "AF_PPPOX",                 /* 24: PPPoX sockets.  */
    "AF_WANPIPE",               /* 25: Wanpipe API sockets.  */
    "",                         /* 26: */
    "",                         /* 27: */
    "",                         /* 28: */
    "",                         /* 29: */
    "",                         /* 30: */
    "AF_BLUETOOTH",             /* 31: Bluetooth sockets.  */
#if 0
    "AF_MAX",                   /* 32: For now..  */
#endif
};

const char * family_sun[] = {
    "AF_UNSPEC",                /* 0: unspecified */
    "AF_UNIX",                  /* 1: local to host (pipes, portals) */
    "AF_INET",                  /* 2: internetwork: UDP, TCP, etc. */
    "AF_IMPLINK",               /* 3: arpanet imp addresses */
    "AF_PUP",                   /* 4: pup protocols: e.g. BSP */
    "AF_CHAOS",                 /* 5: mit CHAOS protocols */
    "AF_NS",                    /* 6: XEROX NS protocols */
    "AF_NBS",                   /* 7: nbs protocols */
    "AF_ECMA",                  /* 8: european computer manufacturers */
    "AF_DATAKIT",               /* 9: datakit protocols */
    "AF_CCITT",                 /* 10: CCITT protocols, X.25 etc */
    "AF_SNA",                   /* 11: IBM SNA */
    "AF_DECnet",                /* 12: DECnet */
    "AF_DLI",                   /* 13: Direct data link interface */
    "AF_LAT",                   /* 14: LAT */
    "AF_HYLINK",                /* 15: NSC Hyperchannel */
    "AF_APPLETALK",             /* 16: Apple Talk */
    "AF_NIT",                   /* 17: Network Interface Tap */
    "AF_802",                   /* 18: IEEE 802.2, also ISO 8802 */
    "AF_OSI",                   /* 19: umbrella for all families used */
    "AF_X25",                   /* 20: CCITT X.25 in particular */
    "AF_OSINET",                /* 21: AFI = 47, IDI = 4 */
    "AF_GOSIP",                 /* 22: U.S. Government OSI */
    "AF_IPX",                   /* 23: Novell Internet Protocol */
    "AF_ROUTE",                 /* 24: Internal Routing Protocol */
    "AF_LINK",                  /* 25: Link-layer interface */
    "AF_INET6",                 /* 26: Internet Protocol, Version 6 */
    "AF_KEY",                   /* 27: Security Association DB socket */
#if 0
    "AF_MAX",                   /* 27: For now..  */
#endif
};

const char *protocol[] = {
    "IP",                       /* 0:  internet protocol, pseudo protocol number */
    "ICMP",                     /* 1:  internet control message protocol */
    "IGMP",                     /* 2:  internet group management protocol */
    "GGP",                      /* 3:  gateway-gateway protocol */
    "IP-ENCAP",                 /* 4:  IP encapsulated in IP (officially ``IP'') */
    "ST",                       /* 5:  ST datagram mode */
    "TCP",                      /* 6:  transmission control protocol */
    "CBT",                      /* 7:  CBT, Tony Ballardie <A.Ballardie@cs.ucl.ac.uk> */
    "EGP",                      /* 8:  exterior gateway protocol */
    "IGP",                      /* 9:  any private interior gateway (Cisco: for IGRP) */
    "BBN-RCC-MON",              /* 10:  BBN RCC Monitoring */
    "NVP-II",                   /* 11:  Network Voice Protocol */
    "PUP",                      /* 12:  PARC universal packet protocol */
    "ARGUS",                    /* 13:  ARGUS */
    "EMCON",                    /* 14:  EMCON */
    "XNET",                     /* 15:  Cross Net Debugger */
    "CHAOS",                    /* 16:  Chaos */
    "UDP",                      /* 17:  user datagram protocol */
    "MUX",                      /* 18:  Multiplexing protocol */
    "DCN-MEAS",                 /* 19:  DCN Measurement Subsystems */
    "HMP",                      /* 20:  host monitoring protocol */
    "PRM",                      /* 21:  packet radio measurement protocol */
    "XNS-IDP",                  /* 22:  Xerox NS IDP */
    "TRUNK-1",                  /* 23:  Trunk-1 */
    "TRUNK-2",                  /* 24:  Trunk-2 */
    "LEAF-1",                   /* 25:  Leaf-1 */
    "LEAF-2",                   /* 26:  Leaf-2 */
    "RDP",                      /* 27:  "reliable datagram" protocol */
    "IRTP",                     /* 28:  Internet Reliable Transaction Protocol */
    "ISO-TP4",                  /* 29:  ISO Transport Protocol Class 4 */
    "NETBLT",                   /* 30:  Bulk Data Transfer Protocol */
    "MFE-NSP",                  /* 31:  MFE Network Services Protocol */
    "MERIT-INP",                /* 32:  MERIT Internodal Protocol */
    "SEP",                      /* 33:  Sequential Exchange Protocol */
    "3PC",                      /* 34:  Third Party Connect Protocol */
    "IDPR",                     /* 35:  Inter-Domain Policy Routing Protocol */
    "XTP",                      /* 36:  Xpress Tranfer Protocol */
    "DDP",                      /* 37:  Datagram Delivery Protocol */
    "IDPR-CMTP",                /* 38:  IDPR Control Message Transport Proto */
    "TP++",                     /* 39:  TP++ Transport Protocol */
    "IL",                       /* 40:  IL Transport Protocol */
    "IPv6",                     /* 41:  IPv6 */
    "SDRP",                     /* 42:  Source Demand Routing Protocol */
    "IPv6-Route",               /* 43:  Routing Header for IPv6 */
    "IPv6-Frag",                /* 44:  Fragment Header for IPv6 */
    "IDRP",                     /* 45:  Inter-Domain Routing Protocol */
    "RSVP",                     /* 46:  Resource ReSerVation Protocol */
    "GRE",                      /* 47:  Generic Routing Encapsulation */
    "MHRP",                     /* 48:  Mobile Host Routing Protocol */
    "BNA",                      /* 49:  BNA */
    "IPv6-Crypt",               /* 50:  Encryption Header for IPv6 */
    "IPv6-Auth",                /* 51:  Authentication Header for IPv6 */
    "I-NLSP",                   /* 52:  Integrated Net Layer Security TUBA */
    "SWIPE",                    /* 53:  IP with Encryption */
    "NARP",                     /* 54:  NBMA Address Resolution Protocol */
    "MOBILE",                   /* 55:  IP Mobility */
    "TLSP",                     /* 56:  Transport Layer Security Protocol */
    "SKIP",                     /* 57:  SKIP */
    "IPv6-ICMP",                /* 58:  ICMP for IPv6 */
    "IPv6-NoNxt",               /* 59:  No Next Header for IPv6 */
    "IPv6-Opts",                /* 60:  Destination Options for IPv6 */
    "",                         /* 61   # any host internal protocol */
    "CFTP",                     /* 62:  CFTP */
    "",                         /* 63   # any local network */
    "SAT-EXPAK",                /* 64:  SATNET and Backroom EXPAK */
    "KRYPTOLAN",                /* 65:  Kryptolan */
    "RVD",                      /* 66:  MIT Remote Virtual Disk Protocol */
    "IPPC",                     /* 67:  Internet Pluribus Packet Core */
    "",                         /* 68   # any distributed file system */
    "SAT-MON",                  /* 69:  SATNET Monitoring */
    "VISA",                     /* 70:  VISA Protocol */
    "IPCV",                     /* 71:  Internet Packet Core Utility */
    "CPNX",                     /* 72:  Computer Protocol Network Executive */
    "CPHB",                     /* 73:  Computer Protocol Heart Beat */
    "WSN",                      /* 74:  Wang Span Network */
    "PVP",                      /* 75:  Packet Video Protocol */
    "BR-SAT-MON",               /* 76:  Backroom SATNET Monitoring */
    "SUN-ND",                   /* 77:  SUN ND PROTOCOL-Temporary */
    "WB-MON",                   /* 78:  WIDEBAND Monitoring */
    "WB-EXPAK",                 /* 79:  WIDEBAND EXPAK */
    "ISO-IP",                   /* 80:  ISO Internet Protocol */
    "VMTP",                     /* 81:  Versatile Message Transport */
    "SECURE-VMTP",              /* 82:  SECURE-VMTP */
    "VINES",                    /* 83:  VINES */
    "TTP",                      /* 84:  TTP */
    "NSFNET-IGP",               /* 85:  NSFNET-IGP */
    "DGP",                      /* 86:  Dissimilar Gateway Protocol */
    "TCF",                      /* 87:  TCF */
    "EIGRP",                    /* 88:  Enhanced Interior Routing Protocol (Cisco) */
    "OSPFIGP",                  /* 89:  Open Shortest Path First IGP */
    "Sprite-RPC",               /* 90:  Sprite RPC Protocol */
    "LARP",                     /* 91:  Locus Address Resolution Protocol */
    "MTP",                      /* 92:  Multicast Transport Protocol */
    "AX.25",                    /* 93:  AX.25 Frames */
    "IPIP",                     /* 94:  Yet Another IP encapsulation */
    "MICP",                     /* 95:  Mobile Internetworking Control Pro. */
    "SCC-SP",                   /* 96:  Semaphore Communications Sec. Pro. */
    "ETHERIP",                  /* 97:  Ethernet-within-IP Encapsulation */
    "ENCAP",                    /* 98:  Yet Another IP encapsulation */
    "",                         /* 99   # any private encryption scheme */
    "GMTP",                     /* 100:  GMTP */
    "IFMP",                     /* 101:  Ipsilon Flow Management Protocol */
    "PNNI",                     /* 102:  PNNI over IP */
    "PIM",                      /* 103:  Protocol Independent Multicast */
    "ARIS",                     /* 104:  ARIS */
    "SCPS",                     /* 105:  SCPS */
    "QNX",                      /* 106:  QNX */
    "A/N",                      /* 107:  Active Networks */
    "IPComp",                   /* 108:  IP Payload Compression Protocol */
    "SNP",                      /* 109:  Sitara Networks Protocol */
    "Compaq-Peer",              /* 110:  Compaq Peer Protocol */
    "IPX-in-IP",                /* 111:  IPX in IP */
    "VRRP",                     /* 112:  Virtual Router Redundancy Protocol */
    "PGM",                      /* 113:  PGM Reliable Transport Protocol */
    "",                         /* 114   # any 0-hop protocol */
    "L2TP",                     /* 115:  Layer Two Tunneling Protocol */
    "DDX",                      /* 116:  D-II Data Exchange */
    "IATP",                     /* 117:  Interactive Agent Transfer Protocol */
    "ST",                       /* 118:  Schedule Transfer */
    "SRP",                      /* 119:  SpectraLink Radio Protocol */
    "UTI",                      /* 120:  UTI */
    "SMP",                      /* 121:  Simple Message Protocol */
    "SM",                       /* 122:  SM */
    "PTP",                      /* 123:  Performance Transparency Protocol */
    "ISIS",                     /* 124:  ISIS over IPv4 */
    "ISIS",                     /* 124:  ISIS over IPv4 */
    "CRTP",                     /* 126:  Combat Radio Transport Protocol */
    "CRUDP",                    /* 127:  Combat Radio User Datagram */
    "CRUDP",                    /* 127:  Combat Radio User Datagram */
    "CRUDP",                    /* 127:  Combat Radio User Datagram */
    "SPS",                      /* 130:  Secure Packet Shield */
    "PIPE",                     /* 131:  Private IP Encapsulation within IP */
    "SCTP",                     /* 132:  Stream Control Transmission Protocol */
    "FC",                       /* 133:  Fibre Channel */
    "",                         /* 134-254   # Unassigned */
    "",                         /* 255   # Reserved */
};

