/*
 *  This file is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  It is part of httppc, which is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/socket.h>     // AF_INET, AF_INET6
#include <netinet/in.h>     // ntohl()
#include <stdio.h>          // snprintf()
#include <errno.h>          // errno
#include <sys/poll.h>       // poll(), POLLIN, POLLOUT

#include "config.h"
#include "internal.h"

// Convert IPv4 address to printable form.
void
ipv4n_to_ipv4p(const struct in_addr *saddr, char *dst, int cnt)
{
    unsigned addr = ntohl(saddr->s_addr);
    snprintf(dst, cnt, "%d.%d.%d.%d", 
             (addr>>24)&0xff, (addr>>16)&0xff, 
             (addr>>8)&0xff, addr&0xff);
}

// Convert IPv6 address to printable and IPv4 compatible form.
#ifdef AF_INET6
void
ipv6n_to_ipv4p(const struct in6_addr *saddr6, char *dst, int cnt)
{
    snprintf(dst, cnt, "%d.%d.%d.%d", 
             saddr6->s6_addr[12], saddr6->s6_addr[13], 
             saddr6->s6_addr[14], saddr6->s6_addr[15]);
}
#endif // AF_INET6

// Convert address in sockaddr structure to printable and IPv4 
// compatible form.
// Returns 0, if family is neither AF_INET, nor AF_INET6.
int
ipn_to_ipv4p(const struct sockaddr *saddr, char *dst, int cnt)
{
    if (saddr->sa_family == AF_INET)
    {
        const struct sockaddr_in *p = (const struct sockaddr_in *)saddr;
        ipv4n_to_ipv4p((const struct in_addr *)&p->sin_addr, dst, cnt);
#ifdef AF_INET6
    } else if (saddr->sa_family == AF_INET6)
    {
        const struct sockaddr_in6 *p = (const struct sockaddr_in6 *)saddr;
        ipv6n_to_ipv4p((const struct in6_addr *)&p->sin6_addr, dst, cnt);
#endif // AF_INET6
    } else
        return 0;
    return 1;
}

int
ipn_to_ipv4n(const struct sockaddr *saddr, struct in_addr *dst)
{
    if (saddr->sa_family == AF_INET)
    {
        const struct sockaddr_in *p = (const struct sockaddr_in *)saddr;
        *dst = p->sin_addr;
#ifdef AF_INET6
    } else if (saddr->sa_family == AF_INET6)
    {
        const struct sockaddr_in6 *p = (const struct sockaddr_in6 *)saddr;
        dst->s_addr = (((((p->sin6_addr.s6_addr[15]<<8)
                          |p->sin6_addr.s6_addr[14])<<8)
                          |p->sin6_addr.s6_addr[13])<<8)
                          |p->sin6_addr.s6_addr[12];
#endif // AF_INET6
    } else
        return 0;
    return 1;
}

/** Wait for specified event on socket.  Handles POLLERR.
    @arg sockfd Socket
    @arg events poll() events
    @arg timeout poll() timeout
    @return -1 in case of error
  */
static int
wait_poll(int sockfd, short events, int timeout)
{
    struct pollfd fds;
    int r;
    
    fds.fd = sockfd;
    fds.events = events | POLLERR;
    fds.revents = 0;
    r = poll(&fds, 1, timeout * 1000);
    return r < 0 ? r : (fds.revents & POLLERR ? (errno = EHOSTUNREACH, -1) : r);
}

int
wait_canread(int sockfd, int timeout)
{
    return wait_poll(sockfd, POLLIN, timeout);
}

int
wait_canwrite(int sockfd, int timeout)
{
    return wait_poll(sockfd, POLLOUT, timeout);
}

int
wait_canrw(int sockfd, int timeout)
{
    return wait_poll(sockfd, POLLIN|POLLOUT, timeout);
}

