/*
 *  This file is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  It is part of httppc, which is
 *    Copyright (C) 2001 Alexey Klimkin <klimkin@mcst.ru>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>     // strtok_r(), strchr(), strncpy()
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>  // inet_pton()
#include <dlfcn.h>      // dlopen(), dlsym()
#include <stdlib.h>     // getenv(), exit()
#include <stdio.h>      // FILE, fopen(), fgets()
#include <errno.h>      // errno
#include <sys/poll.h>   // poll()

#include "config.h"
#include "httppc.h"
#include "internal.h"
#include "sysproto.h"
#include "options.h"

// parses list of networks, and fills net_entry structures
// str should be in format <network>/<mask> .. <network>/<mask>
// i is the first empty net_entry
// returns the number of empty net_entry
static int
parse_nets_list(int i, const char *str, char dst[MAX_NETS][BUFF_LEN], struct net_entry ne[MAX_NETS])
{
    char *tail, *token;
    if (str)
    {
        char buffer[BUFF_LEN];
        strncpy(buffer, str, BUFF_LEN);

        token = strtok_r(buffer, " ", &tail);
        while (token && i < MAX_NETS)
        {
            char *net, *mask, *ntail;

            strncpy(dst[i], token, BUFF_LEN);
            net  = strtok_r(token, "/", &ntail);
            mask = strtok_r(0, "/", &ntail);

            if (net && mask 
                && inet_pton(AF_INET, net, &ne[i].network) > 0
                && inet_pton(AF_INET, mask, &ne[i].mask) > 0)
                i++;
            else
                logerr(PFX"invalid network entry: %s\n", dst[i]);

            token = strtok_r(0, " ", &tail);
        }
    }
    return i;
}

// returns the number of net_entry, if found in list,
// else returns -1
int
in_nets_list(struct in_addr t, struct net_entry *ne, int len)
{
    int i;
    for (i = 0; i < len; i++)
    {
        if ((t.s_addr & ne[i].mask) == ne[i].network)
            return i;
    }
    return -1;
}

// find function funcname in libraries list, specified in format 
// <lib>:<lib>:...:<lib>, returned function pointer is different
// from local pointer
static void *
get_real_entry(const char *libs, const char *funcname, void *local)
{
    void *func = 0, *lib;
    const char *dlerr = 0;
    char llibs[BUFF_LEN], *buffer, *libname;

    strncpy(llibs, libs, BUFF_LEN);
    libname = strtok_r(llibs, ":", &buffer);

    while (libname)
    {
        log(PFX"  %s() in %s: ", funcname, libname);

        lib = dlopen(libname, RTLD_NOW);
        if (!(dlerr = dlerror()) && lib)
        {
            func = dlsym(lib, funcname);

            if (!(dlerr = dlerror()) && func && !(local && func == local))
            {
                log("yes\n");
                return func;
            }
        }

        log("no\n");
        libname = strtok_r(0, ":", &buffer);
    }

    logerr(PFX"fatal: can't find %s() in %s\n", funcname, libs);
    exit(1);
    return 0;
}

// parse options and fill cfg structure
static void
parse_options(struct options *cfg,
              const char *hp, const char *hp_use_lib, const char *hp_mode, 
              const char *x_dns, const char *xnets, const char *lnets, 
              const char *auth, const char *useragent, int timeout,
              httppc_args conn_args)
{
    if (hp_mode && hp_mode[0])
    {
        if (!strcmp(hp_mode, "verbose"))
            httppc__mode = MODE_VERBOSE;
        else if (!strcmp(hp_mode, "quiet"))
            httppc__mode = MODE_QUIET;
    }

    log(PFX"Version "VERSION"\n");
    log(PFX"initializing http tunneling driver...\n");

    log(PFX"  proxy: %s\n", hp ? hp : "");
    log(PFX"  using libs: %s\n", hp_use_lib ? hp_use_lib : "");
    log(PFX"  log mode: %s\n", hp_mode ? hp_mode : "");
    log(PFX"  extern DNS server: %s\n", x_dns ? x_dns : "");
    log(PFX"  external networks: %s\n", xnets ? xnets : "");
    log(PFX"  local networks: %s\n", lnets ? lnets : "");
    log(PFX"  authorization: %s\n", auth ? auth : "");
    log(PFX"  useragent: %s\n", useragent ? useragent : "");
    log(PFX"  timeout: %d\n", timeout);
    log(PFX"  nocache: %s, keepalive: %s, ldns: %s\n",
        conn_args & httppc_nocache   ? "yes" : "no",
        conn_args & httppc_keepalive ? "yes" : "no",
        conn_args & httppc_ldns      ? "yes" : "no");

    if (!(hp_use_lib && hp_use_lib[0]))
    {
        // can't proceed without real_connect
        logerr(PFX"fatal: path to libraries with connect() and gethostbyname() symbols is not specified\n");
        exit(1);
    }

    // retrieve real entries
    cfg->real_connect            = get_real_entry(hp_use_lib, "connect", connect);
    cfg->real_gethostbyname      = get_real_entry(hp_use_lib, "gethostbyname", gethostbyname);
#ifdef HAVE_GETHOSTBYNAME_R
    cfg->real_gethostbyname_r    = get_real_entry(hp_use_lib, "gethostbyname_r", gethostbyname_r);
#endif // HAVE_GETHOSTBYNAME_R
    cfg->real_gethostbyaddr      = get_real_entry(hp_use_lib, "gethostbyaddr", gethostbyaddr);
#ifdef HAVE_GETHOSTBYADDR_R
    cfg->real_gethostbyaddr_r    = get_real_entry(hp_use_lib, "gethostbyaddr_r", gethostbyaddr_r);
#endif // HAVE_GETHOSTBYADDR_R
    cfg->real_socket             = get_real_entry(hp_use_lib, "socket", socket);
#if 0
    cfg->real_select             = get_real_entry(hp_use_lib, "select", select);
    cfg->real_poll               = get_real_entry(hp_use_lib, "poll", poll);
#endif
#ifdef HAVE_GETADDRINFO
    cfg->real_getaddrinfo        = get_real_entry(hp_use_lib, "getaddrinfo", getaddrinfo);
#endif // HAVE_GETADDRINFO

    // --
    if (hp && hp[0])
    {
        char *tail, *token;
        char buffer[BUFF_LEN];

        strncpy(buffer, hp, BUFF_LEN);
        // parse proxy string
        token = strtok_r(buffer, ":", &tail);
        if (token)
        {
            strncpy(cfg->http_proxy, token, BUFF_LEN);
            token = strtok_r(0, ":", &tail);
            if (token)
                cfg->http_proxy_port = atoi(token);
            else
            {
                log(PFX"  using default proxy port %d\n", DFLT_PROXY_PORT);
                cfg->http_proxy_port = DFLT_PROXY_PORT;
            }
        } else
        {
            logerr(PFX"fatal: specify proxy in form <address>[:<port>]\n");
            exit(1);
        }
    }

    // --
    if (x_dns && x_dns[0])
        strncpy(cfg->extern_dns, x_dns, BUFF_LEN);
    else
    {
        log(PFX"  external DNS is not specified, using default: "DFLT_XDNS_NAME" ("DFLT_XDNS")\n");
        strncpy(cfg->extern_dns, DFLT_XDNS, BUFF_LEN);
    }
    
    // --
    if (useragent && useragent[0])
        strncpy(cfg->useragent, useragent, BUFF_LEN);

    // --
    cfg->extern_lsz = parse_nets_list(0, xnets, cfg->extern_nets, cfg->extern_ne);
    cfg->local_lsz  = parse_nets_list(0, lnets, cfg->local_nets, cfg->local_ne);

    // --
    if (auth && auth[0])
    {
        char *tail, *token;
        char buffer[BUFF_LEN];

        strncpy(buffer, auth, BUFF_LEN);
        // parse authorization string
        token = strtok_r(buffer, ":", &tail);
        if (token)
        {
            strncpy(cfg->proxy_user, token, BUFF_LEN);
            token = strtok_r(0, ":", &tail);
            if (token)
                strncpy(cfg->proxy_pass, token, BUFF_LEN);
            else
            {
                logerr(PFX"fatal: specify authorization in form <user>:<password>\n");
                exit(1);
            }
        } else
        {
            logerr(PFX"fatal: specify authorization in form <user>:<password>\n");
            exit(1);
        }
    }

    // --
    cfg->timeout = timeout;
    // --
    cfg->conn_args = conn_args;
}

// load parameters from file nm
static int
load_conffile(const char *nm,
              char *hp, char *hp_use_lib, char *hp_mode, 
              char *x_dns, char *xnets, char *lnets, 
              char *auth, char *useragent, int *timeout,
              httppc_args *conn_args)
{
    FILE *fp;
    char line[BUFF_LEN];

    if (!(fp = fopen(nm, "r")))
        return 0;

    while(fgets(line, BUFF_LEN, fp))
    {
        char *token, *tail;
        
        if ((token = strchr(line, '#')))    // remove comment
            token[0] = '\0';
        if ((token = strchr(line, '\012'))) // remove newline
            token[0] = '\0';
        if ((token = strchr(line, '\010'))) // remove newline
            token[0] = '\0';

        token = strtok_r(line, " \t", &tail);
        if (!token)
            continue;
       
        while (tail && (*tail == ' ' || *tail == '\t')) // cat leading whitespaces
            tail++;

        if (!strcmp("proxy", token))
            strncpy(hp, tail, BUFF_LEN);
        else if (!strcmp("log", token))
            strncpy(hp_mode, tail, BUFF_LEN);
        else if (!strcmp("libs", token))
            strncpy(hp_use_lib, tail, BUFF_LEN);
        else if (!strcmp("xdns", token))
            strncpy(x_dns, tail, BUFF_LEN);
        else if (!strcmp("local", token))
            strncpy(lnets, tail, BUFF_LEN);
        else if (!strcmp("extern", token))
            strncpy(xnets, tail, BUFF_LEN);
        else if (!strcmp("nocache", token))
            *conn_args |= httppc_nocache;
        else if (!strcmp("keepalive", token))
            *conn_args |= httppc_keepalive;
        else if (!strcmp("ldns", token))
            *conn_args |= httppc_ldns;
        else if (!strcmp("auth", token))
        {
            *conn_args |= httppc_auth;
            strncpy(auth, tail, BUFF_LEN);
        } else if (!strcmp("useragent", token))
        {
            *conn_args |= httppc_useragent;
            strncpy(useragent, tail, BUFF_LEN);
        } else if (!strcmp("timeout", token))
        {
            *timeout = atoi(tail);
        } else 
            logerr(PFX"unknown option in `%s': %s\n", nm, token);
    }
    return 1;
}

// first run initialization
void
init(int *initialized, struct options *cfg)
{
    char *home;
    char hp[BUFF_LEN], hp_use_lib[BUFF_LEN], hp_mode[BUFF_LEN], 
         x_dns[BUFF_LEN], xnets[BUFF_LEN], lnets[BUFF_LEN], 
         auth[BUFF_LEN], useragent[BUFF_LEN];
    int timeout = DEFAULT_TIMEOUT;
    char *env_hp, *env_hp_use_lib, *env_hp_mode, 
         *env_x_dns, *env_xnets, *env_lnets,
         *env_auth, *env_useragent, *env_timeout;
    httppc_args conn_args;

    // do all initialization here
    if (*initialized)
        return;
    *initialized = 1;

    hp[0] = hp_use_lib[0] = hp_mode[0] = x_dns[0] = xnets[0] = lnets[0] = auth[0] = useragent[0] = '\0';
    conn_args = httppc_none;
    memset(cfg, 0, sizeof(*cfg));

    load_conffile(CONF_SYS_WIDE, hp, hp_use_lib, hp_mode, x_dns, xnets, lnets, 
                  auth, useragent, &timeout, &conn_args);
    home = getenv("HOME");
    if (home && home[0])
    {
        char buffer[BUFF_LEN];
        snprintf(buffer, BUFF_LEN, "%s/"CONF_USER, home);
        load_conffile(buffer, hp, hp_use_lib, hp_mode, x_dns, xnets, lnets, 
                      auth, useragent, &timeout, &conn_args);
    }

    env_hp          = getenv("HTTP_PROXY");
    env_hp_use_lib  = getenv("HTTP_PROXY_LIBS");
    env_hp_mode     = getenv("HTTP_PROXY_LOG");
    env_x_dns       = getenv("HTTP_PROXY_XDNS");
    env_lnets       = getenv("HTTP_PROXY_LOCAL");
    env_xnets       = getenv("HTTP_PROXY_EXTERN");
    env_auth        = getenv("HTTP_PROXY_AUTH");
    env_useragent   = getenv("HTTP_PROXY_USERAGENT");
    env_timeout     = getenv("HTTP_PROXY_TIMEOUT");

    parse_options(cfg, 
                  env_hp && env_hp[0]                   ? env_hp : hp,
                  env_hp_use_lib && env_hp_use_lib[0]   ? env_hp_use_lib : hp_use_lib,
                  env_hp_mode && env_hp_mode[0]         ? env_hp_mode : hp_mode,
                  env_x_dns && env_x_dns[0]             ? env_x_dns : x_dns,
                  env_xnets && env_xnets[0]             ? env_xnets : xnets,
                  env_lnets && env_lnets[0]             ? env_lnets : lnets,
                  env_auth && env_auth[0]               ? env_auth : auth,
                  env_useragent && env_useragent[0]     ? env_useragent : useragent,
                  env_timeout && env_timeout[0]         ? atoi(env_timeout) : timeout,
                  conn_args
                  );

    // --
    log(PFX"initialization completed\n");
}

